/*
 * Adds username and password to ssmtp.config, encoded to base64 so that
 * they aren't in the clear where anyone can just read them.
 */

#include "ssmtp.h"
char *strncasestr (char *, char *, size_t);
void onexit (void);

struct	configlines {
		struct 	configlines	*next;
		char			*line;
}
	*config = NULL;

int main (int argc, char **argv) {

  FILE *fp = NULL;
  char	line[MAXLINE],
  	*p,
  	*username = NULL,
  	*password = NULL,
  	*file = NULL,
  	*progname = argv[0];
  int i;
  struct configlines	*thisline = NULL,
  			*lastline = NULL;

  /* get options */
  if (argc < 2 || argc > 4) {
    fprintf (stderr, "usage: %s -uUsername -pPassword [configfile]\n", progname);
    return 1;
  }
  for (i=1;i<argc;++i) {
    switch (argv[i][0]) {
      case '-':
        switch (argv[i][1]) {
          case 'u':
            if (!argv[i][2]) {
              fprintf (stderr,
       	        "usage: %s -uUsername -pPassword [configfile]\n", progname);
              return 1;
            }
            username = &argv[i][2];
            break;
          case 'p':
            if (!argv[i][2]) {
              fprintf (stderr,
                "usage: %s -uUsername -pPassword [configfile]\n", progname);
              return 1;
            }
            password = &argv[i][2];
            break;
          default:
            fprintf (stderr, "%s: unrecognized option %s\n", progname, argv[i]);
            return 1;
        }
        break;
      default:
        file = argv[i];
        break;
    }
  }
  if (!username && !password) {
    fprintf (stderr, "%s: neither username nor password specified. nothing to do!\n",
      progname);
    return 1;
  }

  /* open config file */
  if (file && !(fp = fopen (file, "r"))) {
    fprintf (stderr, "%s: unable to open %s\n", progname, file);
    return 1;
  }
  else {
    char	*locations[] =
    		{
    			CONFIGURATION_FILE,
    			NULL
    		},
    		**lp;

    for (lp = &locations[0]; *lp; lp++) {
      if ((fp = fopen (*lp, "r"))) {
        file = *lp;
        break;
      }
    }
    if (!fp) {
      fprintf (stderr, "%s: unable to open configuration file\n", progname);
      return 1;
    }
  }

  /* read config file */
  atexit (onexit);
  while (fgets (line, MAXLINE, fp)) {
    /* leave out existing 'username' and 'password' entries */
    if ((p = strchr (line, '=')) &&
         (strncasestr (line, "username", p - line) ||
          strncasestr (line, "password", p - line)))
            continue;
    if (!(thisline = malloc (sizeof *thisline))) {
      fprintf (stderr, "%s: out of memory\n", progname);
      return 1;
    }
    thisline->next = NULL;
    if (!(thisline->line = strdup (line))) {
      fprintf (stderr, "%s: out of memory\n", progname);
      return 1;
    }
    if (config)
      lastline->next = thisline;
    else
      config = thisline;
    lastline = thisline;
  }
  (void) fclose (fp);

  /* write it back out again, add new 'username' and 'password' entries */
  if (!(fp = fopen (file, "wb"))) {
    fprintf (stderr, "%s: can't open %s for writing new config\n",
      progname, file);
    return 1;
  }
  for (thisline = config; thisline; thisline = thisline->next)
    fprintf (fp, "%s", thisline->line);
  to64frombits (line, username, strlen (username));
  fprintf (fp, "Username=%s\n", line);
  to64frombits (line, password, strlen (password));
  fprintf (fp, "Password=%s\n", line);
  fclose (fp);
  if (username && password) {
    printf ("username and password successfully added to %s\n", file);
  }
  else {
    if (username) {
      printf ("username ");
    }
    else if (password) {
      printf ("password ");
    }
    printf ("successfully added to %s\n", file);
  }
  return 0;
}


/*
 * counted case-insensitive search for string s2 in string s1
 */
char *strncasestr(char *s1, char *s2, size_t n) {

	char *p, *q, *r;

	if (!s1 || !s2 || !*s2 || strlen (s2) > n) return NULL;
	for (p = s1; *p && p - s1 < n; p++) {
		if (toupper(*p) == toupper(*s2)) {
			q=p+1; r=s2+1;
			while (*q && *r && toupper(*q++) == toupper(*r++));
			if (!*r) return p;
		}
	}
	return NULL;
}

void onexit (void) {

  struct configlines *p, *np;
  
  for (p = config; p; p = np) {
    np = p->next;
    free (p->line);
    free (p);
  }
}
