/*
 * sSMTP sendmail -- send messages via smtp to a mailhub for local delivery 
 *      or forwarding. This program is used in place of /usr/lib/sendmail,
 *      called by /bin/mail (et all).   sSMTP does a selected subset of 
 *      sendmail's standard tasks (including exactly one rewriting task), and
 *      explains if you ask it to do something it can't.  It then sends
 *      the mail to the mailhub via an smtp connection.  Believe it or not,
 *      this is nothing but a filter.  You can receive mail with inetd, an
 *      inverse filter and /bin/mail -d.
 *
 * See COPYRIGHT for the license.
 * 
 * See CHANGELOG for the list of changes.
 * 
 */

/*
 * Patched up version. Start by organizing the code-- #includes, #defines, 
 * global variable declarations, and function prototypes collected together
 * in ssmtp.h, and a prototype there for each function here. Code from
 * arpadate.c merged into ssmtp.[ch].  -Jeff
 */

#include "ssmtp.h"

static int	hasFrom = NO,
		hasDSN = NO,
#ifdef HASTO_OPTION
		hasTo = NO,
#endif
		hasDate = NO;

/*
 * main -- make the program behave like sendmail, then call ssmtp.
 */
int 
main (int argc, char **argv)
{
  char **newArgv, buffer[MAXLINE], *p;
  int fd, i;
  struct recips	*thisheader = NULL,
  		*lastheader = NULL;

  /* Try to be bulletproof (:-)) */
  (void) signal (SIGHUP, SIG_IGN);
  (void) signal (SIGINT, SIG_IGN);
  (void) signal (SIGTTIN, SIG_IGN);
  (void) signal (SIGTTOU, SIG_IGN);

  /* Set the globals. */
  ProgName = argv[0];
  if (gethostname (HostName, sizeof (HostName)) == ERR)
    {
      die ("can't find the name of this host, %s, exiting.",
	   "(an impossible condition)");
    }
  if ((Sender = getpwuid (getuid ())) == NULL)
    {
      die ("couldn't find password entry for sender (uid %d).",
	   getuid ());
    }
  (void) get_arpadate (DateString);

  newArgv = doOptions (argc, argv);
#if 0  
  printf("authUsername=%s  authPassword=%s\n",authUsername,authPassword);
#endif  

/*
 * ssmtp -- send the message (exactly one) from stdin to the smtp
 *      port on the mailhub.
 *
 * 	(This is just a continuation of main()-- might as well inline it
 *	and save a frame on the stack.)  -Jeff
 */

  if (getConfig () == NO)
    {
      log_event (LOG_INFO, "No ssmtp.conf in %s", SSMTPCONFDIR);
    }
  p = strtok (Sender->pw_gecos, ";,");
  if (p)
    fullName = strdup (p);
  getReverseAliases ();
  if (!fromLine)  /* no reverse alias for sender */
    {  
      localFromLine ();
    }

  /* Header analysis first */
  /* allocation algorithm for header buffer missing */
  /* not anymore.. :) -Jeff */
  while (fgets (buffer, sizeof buffer, stdin) && (*buffer != '\n'))
    {
      if (Verbose)
	fprintf (stderr, "Read Header: %s", buffer);
      /* Trim off (\r)?\n, double leading .'s */
      standardize (buffer);
      if (recordRequiredHeaders (buffer)) continue;
      fixFromLines (buffer);
      if (!(thisheader = malloc (sizeof *thisheader))) die ("out of memory\n");
      thisheader->next = NULL;
      if (!(thisheader->rec = strdup (buffer))) die ("out of memory\n");
      if (headers)
        lastheader->next = thisheader;
      else
        headers = thisheader;
      lastheader = thisheader;
    }

  /* Now to the delivery of the message */
  (void) signal (SIGALRM, (void (*)()) alarmHandler);	/* Catch SIGALRMs. */
  (void) alarm ((unsigned) MAXWAIT);	/* Set initial timer. */
  if (setjmp (TimeoutJmpBuf) != 0)
    {
      /* Then the timer has gone off and we bail out. */
      die ("connection lost in middle of processing, exiting.");
    }
  if ((fd = openSocket (MailHub, PortNum)) == ERR)
    {
      die ("can't open the smtp port (%d) on %s.",
	   PortNum, MailHub);
    }
  else if (getOkFromSmtp (fd, buffer) == NO)
    {
      die ("didn't get initial OK message from smtp server.");
    }
  if (Verbose)
    {
      (void) fprintf (stderr, "Connected to smtp server %s\n", MailHub);
    }
  if (LogLevel > 0)
    {
      log_event (LOG_INFO, "Connected to smtp server %s\n", MailHub);
    }

  if (hasDSN) {
  	hasDSN = NO;
  	putToSmtp (fd, "EHLO %s", HostName);
  	(void) alarm ((unsigned) MEDWAIT);
  	do {
  		getLine (buffer, MAXLINE, fd);
  		if (!strncasecmp (buffer + 4, "DSN", 3))
  			hasDSN = YES;
  	} while (buffer[3] == '-');
  	if (atoi (buffer) / 100 != 2)
		die ("server didn't accept hostname %s, replied \"%s\".",
			HostName, buffer);
	if (hasDSN == NO) {
		fprintf (stderr, "%s: mailhub does not support Delivery Status Notification\n",
			ProgName);
	} else {
		fprintf (stderr, "%s: Return Receipt Requested\n", ProgName);
	}
  }
  else {
    if (authUsername) {
      putToSmtp (fd, "EHLO %s", HostName);
    } else {
      putToSmtp (fd, "HELO %s", HostName);
    }
    (void) alarm ((unsigned) MEDWAIT);
    if (getOkFromSmtp (fd, buffer) == NO)
      {
        die ("server didn't accept hostname %s, replied \"%s\".",
  	   HostName, buffer);
      }
  }

  /* try to log in if username was supplied */
  if (authUsername)
    {
      if (coded_auth) {
        putToSmtp (fd, "AUTH LOGIN %s", authUsername);
        }
      else {
        memset(buffer, 0, MAXLINE);
        to64frombits(buffer, authUsername, strlen(authUsername));
        putToSmtp (fd, "AUTH LOGIN %s", buffer);
      }
      (void) alarm ((unsigned) MEDWAIT);
      if (getFromSmtp (fd, buffer) != 3)
        {
	  die ("server didn't accept -auUsername: \"%s\".", buffer);
	}
      if (coded_auth)
        putToSmtp (fd, "%s", authPassword);
      else {
        memset(buffer, 0, MAXLINE);
        to64frombits(buffer, authUsername, strlen(authPassword));
        putToSmtp (fd, "%s", buffer);
      }
      (void) alarm ((unsigned) MEDWAIT);
      if (getOkFromSmtp (fd, buffer) == NO)
        {
	  die ("server didn't accept -apPassword: \"%s\".", buffer);
	}
    }
  
  /* Send "MAIL FROM:" line */
  if (authUsername && !specifiedFrom)
    {
      if (coded_auth) {
        /* decode username, null-terminate string */
        *(from64tobits (buffer, authUsername) + buffer) = 0;
        putToSmtp (fd, "MAIL FROM:<%s@%s>%s", buffer, HostName,
        	hasDSN ? " RET=FULL" : "");
        memset (buffer, 0, sizeof buffer);
      }
      else {
        putToSmtp (fd, "MAIL FROM:<%s@%s>%s", authUsername, HostName,
        	hasDSN ? " RET=FULL" : "");
        if (msgFromLine) free (msgFromLine);
      }
    }
  else if (msgFromLine && FromLineOverride)
    {
      putToSmtp (fd, "MAIL FROM:<%s>%s", stripFromLine(msgFromLine),
      	hasDSN ? " RET=FULL" : "");
      free(msgFromLine);
    }
  else
    putToSmtp (fd, "MAIL FROM:<%s>%s",
      specifiedFrom? specifiedFrom : stripFromLine(fromLine),
      hasDSN ? " RET=FULL" : "");

  (void) alarm ((unsigned) MEDWAIT);
  if (getOkFromSmtp (fd, buffer) == NO)
    {
      die ("smtp server didn't accept MAIL From, replied \"%s\".",
	   buffer);
    }
  if (Verbose)
    {
      (void) fprintf (stderr, "Server accepted MAIL FROM: %s line.\n",
        specifiedFrom? specifiedFrom : stripFromLine(fromLine));
    }
  if (LogLevel > 0)
    {
      log_event (LOG_INFO, "Server accepted MAIL FROM: %s line.\n",
        specifiedFrom? specifiedFrom : stripFromLine(fromLine));
    }

  /* Send all the To: adresses. */
  /* Either we're using the -t option, or we're using the arguments */
  if (Toption)
    {
      struct recips *recip_ent, *np;
      if (!recipients)
	die ("No recipient specified although -t option specified");
      for (recip_ent = recipients; recip_ent; recip_ent = np)
	{
	  strcpy (buffer, properRecipient (recip_ent->rec));
	  putToSmtp (fd, "RCPT TO:<%s>%s%s", buffer,
	  	hasDSN ? " NOTIFY=SUCCESS,FAILURE,DELAY ORCPT=rfc822;" : "",
	  	hasDSN ? buffer : "");
	  (void) alarm ((unsigned) MEDWAIT);
	  if (getOkFromSmtp (fd, buffer) == NO)
	    {
	      die ("smtp server didn't accept RCPT To: command, replied \"%s\".",
		   buffer);
	    }
          np = recip_ent->next;
          free (recip_ent->rec);
          free (recip_ent);
	}
    }
  else
    {
      for (i = 1; newArgv[i] != NULL; i++)
	{
	  p = strtok (newArgv[i], ",");
	  while (p)
	    {
	      /* RFC822 Address  -> "foo@bar" */
              parseaddr(p, buffer, sizeof(buffer));
	      putToSmtp (fd, "RCPT TO:<%s>%s%s", properRecipient (buffer),
	  	hasDSN ? " NOTIFY=SUCCESS,FAILURE,DELAY ORCPT=rfc822;" : "",
	  	hasDSN ? properRecipient (buffer) : "");
	      (void) alarm ((unsigned) MEDWAIT);
	      if (getOkFromSmtp (fd, buffer) == NO)
		{
		  die ("smtp server didn't accept RCPT To: command, replied \"%s\".",
		       buffer);
		}
	      p = strtok (NULL, ",");
	    }
	}
    }
  if (Verbose)
    {
      (void) fprintf (stderr, "Server accepted To: line(s).\n");
    }
  if (LogLevel > 0)
    {
      log_event (LOG_INFO, "Server accepted To: line(s).\n");
    }

  /* Send DATA. */
  putToSmtp (fd, "DATA");
  (void) alarm ((unsigned) MEDWAIT);
  if (getFromSmtp (fd, buffer) != 3)
    {
      /* Oops, we were expecting "354 send your data". */
      die ("smtp server didn't accept DATA, replied \"%s\".",
	   buffer);
    }
  if (Verbose)
    {
      (void) fprintf (stderr, "Message body transmission started.\n");
    }

  /* Send headers, with optional From: rewriting. */
  addInitialHeaders (fd);
  addRequiredHeaders (fd);
  for (thisheader = headers; thisheader; thisheader = lastheader)
    {
      putToSmtp (fd, "%s", thisheader->rec);
      (void) alarm ((unsigned) MEDWAIT);
      lastheader = thisheader->next;
      free (thisheader->rec);
      free (thisheader);
    }
  /* End of headers, start body. */
  putToSmtp (fd, "");		/* Seperate headers and body. */
  while (fgets (buffer, sizeof buffer, stdin) != NULL)
    {
      /* Trim off \n, double leading .'s */
      standardize (buffer);
      putToSmtp (fd, "%s", buffer);
      (void) alarm ((unsigned) MEDWAIT);
    }
  /* End of body. */
  putToSmtp (fd, ".");

  (void) alarm ((unsigned) MAXWAIT);
  if (getOkFromSmtp (fd, buffer) == NO)
    {
      die ("smtp server wouldn't accept message, replied \"%s\".",
	   buffer);
    }
  if (Verbose)
    {
      (void) fprintf (stderr, "Message body transmission complete.\n");
    }
  /* Close conection. */
  (void) signal (SIGALRM, SIG_IGN);
  putToSmtp (fd, "QUIT");
  (void) getOkFromSmtp (fd, buffer);
  (void) close (fd);
  (void) log_event (LOG_INFO, "%s sent mail for %s",
		    ProgName, Sender->pw_name);
  return 0;
}


/*
 * doOptions -- pull the options out of the command-line, process them 
 *      (and special-case calls to mailq, etc), and return the rest.
 */
char **
doOptions (int argc, char **argv)
{
  int i, j, add, newArgC;
  static char *newArgV[MAXARGS];

  newArgV[0] = argv[0];
  newArgC = 1;

  if (strstr (argv[0], "mailq") != NULL)
    {
      /* Someone wants to know the queue state... */
      (void) printf ("Mail queue is empty.\n");
      exit (0);
    }
  else if (strstr (argv[0], "newalias") != NULL)
    {
      /* Someone wanted to recompile aliases. */
      /* This is slightly more like to be a human... */
      die ("newalias is meaningless to sSMTP: it doesn't do aliases.");
    }

  i = 1;
  while (i < argc)
  {
      if (argv[i][0] != '-')
	  {
	  newArgV[newArgC++] = argv[i++];
	  continue;
	  }
      j = 0;
      add = 1;
      while (argv[i][++j] != '\0')
      {
#ifdef DEBUG
	  printf("%c", argv[i][j]);
#endif
	  switch (argv[i][j])
	    {
	    case 'a':
	      coded_auth = NO;
	      switch (argv[i][++j])
	        {
		 case 'u':
		  if (!argv[i][j+1]) 
		  {
		    authUsername = strdup(argv[i+1]);
		    add++;
		  }
		  else
		    authUsername = strdup(argv[i]+j+1);
		  goto exit;
		 case 'p':
		  if (!argv[i][j+1]) 
		  {
		    authPassword = strdup(argv[i+1]);
		    add++;
		  }
		  else
		    authPassword = strdup(argv[i]+j+1);
		  goto exit;
		}
	      goto exit;

	    case 'b':
	      switch (argv[i][++j])
		{
		case 'a':		/* ARPANET mode. */
		  die ("-ba is not supported by sSMTP sendmail, nor is -t.");
		case 'd':		/* Run as a daemon. */
		  die ("-bd is not supported by sSMTP sendmail. Use rSMTP under inetd instead.");
		case 'i':		/* Initialize aliases. */
		  continue;
		case 'm':		/*  Default addr processing. */
		  continue;
		case 'p':		/* Print mailqueue. */
		  die ("Mail queue is empty.");
		case 's':		/* Read smtp from stdin. */
		  die ("-bs is not supported by sSMTP sendmail.");
		case 't':		/* Test mode. */
		  die ("-bt is meaningless to sSMTP sendmail. It doesn't route.");
		case 'v':		/*  Verify names only. */
		  die ("-bv is meaningless to sSMTP sendmail. It doesn't route.");
		case 'z':		/* Create  freeze file. */
		  die ("-bz is meaningless to sSMTP sendmail. It isn't programmable.");
		}
	    case 'C':		/* Configfile name. */
	      goto exit;
	    case 'd':		/* Debug. */
	      LogLevel = 1;
	      Verbose = YES;	/* Almost the same thing... */
	      continue;
	    case 'E':		/* insecure channel, don't trust userid. */
	      continue;
	    case 'R':
	      if (!argv[i][j+1]) {	/* amount of the message to be returned */
		add++;
		goto exit;
	      }
	      else {		/* Process queue for recipient. */
		continue;
	      }
	    case 'F':		/* fullname of sender. */
	      if (!argv[i][j+1]) {
		specifiedName = strdup(argv[i+1]);
		add++;
	      }
	      else {
		specifiedName = strdup(argv[i]+j+1);
	      }
	      goto exit;
	    case 'f':		/* Set from/sender address. */
	    case 'r':		/* Obsolete -f flag. */
	      if (!argv[i][j+1]) {
		specifiedFrom = strdup(argv[i+1]);
		add++;
	      }
	      else {
		specifiedFrom = strdup(argv[i]+j+1);
	      }
	      strcpy(specifiedFrom, stripFromLine(specifiedFrom));
	      goto exit;
	      /* Should I support these??? When? */
	    case 'h':		/* Set hopcount. */
	      continue;
	    case 'm':		/* Ignore originator in adress list. */
	      continue;
	    case 'M':		/* Use specified message-id. */
	      goto exit;
	    case 'N':		/* dsn options */
	      add++;
	      goto exit;
	    case 'n':		/* No aliasing. */
	      continue;
	    case 'o':
	      switch (argv[i][++j])
		{
		case 'A':		/* Alternate aliases file. */
		  goto exit;
		case 'c':		/* Delay connections. */
		  continue;
		case 'D':		/* Run newaliases if rqd. */
		  continue;
		case 'd':		/* Deliver now, in background or queue. */
		  /* This may warrant a diagnostic for b or q. */
		  continue;
		case 'e':		/* Errors: mail, write or none. */
		  continue;
		case 'F':		/* Set tempfile mode. */
		  goto exit;
		case 'f':		/* Save ``From ' lines. */
		  continue;
		case 'g':		/* Set group id. */
		  goto exit;
		case 'H':		/* Helpfile name. */
		  continue;
		case 'i':		/* DATA ends at EOF, not \n.\n */
		  continue;
		case 'L':		/* Log level. */
		  goto exit;
		case 'm':		/* Send to me if in the list. */
		  continue;
		case 'o':		/* Old headers, spaces between adresses. */
		  die ("-oo (old header format) is not supported by sSMTP sendmail.");
		case 'Q':		/* Queue dir. */
		  goto exit;
		case 'r':		/* Read timeout. */
		  goto exit;
		case 's':		/* Always init the queue. */
		  continue;
		case 'S':		/* Stats file. */
		  goto exit;
		case 'T':		/* Queue timeout. */
		  goto exit;
		case 't':		/* Set timezone. */
		  goto exit;
		case 'u':		/* Set uid. */
		  goto exit;
		case 'v':		/* Set verbose flag. */
		  Verbose = YES;
		  continue;
		}
	      break;
	    case 'q':		/* Process the queue [at time] */
	      die ("Mail queue is empty.");
	    case 't':		/* Read message's To/Cc/Bcc lines. */
	      Toption = YES;
	      continue;
	    case 'v':		/* Verbose (ditto -ov). */
	      Verbose = YES;
	      break;
	    case 'V':		/*  Say version and quit. */
	      /* Similar as die, but no logging */
	      fprintf (stderr,
		       "sSMTP version %s (not sendmail at all)\n",
		       Version);
	      flush ();		/* Send message to dead.letter */
	      exit (0);
	      break;
	    }
	}
    exit:
      i += add;
#ifdef DEBUG
      printf("\n");
#endif
  }
  newArgV[newArgC] = NULL;
  if (newArgC <= 1 && !Toption)
    {
      die ("no recipients supplied: no mail will be sent.");
    }
  if (newArgC > 1 && Toption)
    die ("Recipientlist with -t option not supported.");
  return &newArgV[0];
}

/*
 * Config file access and processing routines.
 */
int 
getConfig (void)
{
  FILE *fp;
  static char *locations[] =
  {
    CONFIGURATION_FILE,
    NULL
  };
  char **lp;

  for (lp = &locations[0]; *lp; lp++)
    {
      if ((fp = fopen (*lp, "r")) != NULL)
	{
	  parseConfig (fp);
	  (void) fclose (fp);
	  return YES;
	}
    }
  return NO;			/* We use the default for everything. */
}

/*
 * parseConfig -- parse config file, extract values of a few
 *      predefined variables.
 */
void 
parseConfig (FILE * fp)
{
  char line[MAXLINE], *p, *q, *r;

  while (fgets (line, sizeof line, fp))
    {
      /* Make comments invisible. */
      if ((p = strchr (line, '#')))
	*p = '\0';
      /* Ignore malformed lines and comments. */
      if (strchr (line, '=') == NULL)
	continue;
      /* Parse out keywords. */
      if (((p = strtok (line, "= \t\n")) != NULL)
	  && ((q = strtok (NULL, " \t\n:")) != NULL))
	{
	  if (strcasecmp (p, "Root") == 0)
	    {
	      Root = strdup (q);
	      if (LogLevel > 0)
		{
		  log_event (LOG_INFO,
			     "Set Root=\"%s\".\n", Root);
		}
	    }
	  else if (strcasecmp (p, "MailHub") == 0)
	    {
	      MailHub = strdup (q);
	      PortNum = 0;
              if ((r = strtok (NULL, "= \t\n:")) != NULL) PortNum = atoi(r);
	      if (PortNum == 0) PortNum = PORTNUMBER;
	      if (LogLevel > 0)
		{
		  log_event (LOG_INFO,
			     "Set MailHub=\"%s\".\n", MailHub);
		  log_event (LOG_INFO,
			     "via SMTP Port Number=\"%d\".\n", PortNum);
                  
		}
	    }
	  else if (strcasecmp (p, "HostName") == 0)
	    {
	      (void) strcpy (HostName, q);
	      if (LogLevel > 0)
		{
		  log_event (LOG_INFO,
			     "Set HostName=\"%s\".\n", HostName);
		}
	    }

#ifdef REWRITE_DOMAIN
	  else if (strcasecmp (p, "rewriteDomain") == 0)
	    {
	      RewriteDomain = strdup (q);
	      UseRD = YES;
	      if (LogLevel > 0)
		{
		  log_event (LOG_INFO,
			     "Set RewriteDomain=\"%s\".\n",
			     RewriteDomain);
		}
	    }
#endif
          else if (strcasecmp (p, "FromLineOverride") == 0)
            {
              if (strcasecmp(q, "yes") == 0)
                FromLineOverride = YES;
              else
                FromLineOverride = NO;
              if (LogLevel > 0)
                {
                  log_event (LOG_INFO,
                             "Set FromLineOverride=\"%s\".\n",
                             FromLineOverride ? "YES" : "NO");
                }
            }
/* Add authentication */
	  else if (strcasecmp (p, "Username") == 0
	  	  && !authUsername) {
	  	  	authUsername = strdup (q);
	  	  	coded_auth = YES;
	  }
	  else if (strcasecmp (p, "Password") == 0
	  	  && !authPassword) {
	  	  	authPassword = strdup (q);
	  	  	coded_auth = YES;
	  }
	  else
	    {
	      log_event (LOG_INFO,
			 "Unable to set %s=\"%s\".\n",
			 p, q);
	    }
	}
    }
  return;
}

/*
 * getReverseAliases -- parse the reverse alias file, fix globals to use
 * any entry for sender.
 */
void
getReverseAliases (void)
{
  FILE *rev_file;
  static char buffer[MAXLINE];
  char line[MAXLINE], *p, *r;
  /* Try to open the reverse aliases file */
  rev_file = fopen (REVALIASES_FILE, "r");
  /* Search if a reverse alias is definied for the sender */
  if (rev_file)
    {
      while (fgets (line, sizeof line, rev_file))
	{
	  /* Make comments invisible. */
	  if ((p = strchr (line, '#')))
	    {
	      *p = '\0';
	    }
	  /* Ignore malformed lines and comments. */
	  if (*line == '#' || strchr (line, ':') == NULL)
	    continue;
	  /* Parse the alias */
	  if (((p = strtok (line, ":")) != NULL)
	      && !strcmp (p, Sender->pw_name))
	    {
	      p = strtok (NULL, ": \t\r\n");
	      if (p)
		{
		  if (fullName) {
		    (void) sprintf (buffer, "\"%s\" <%s>", fullName, p);
		    fromLine = strdup (buffer);
		  }
		  else
		    fromLine = strdup (p);
		}
	      p = strtok (NULL, " \t\r\n:");
	      if (p)
	        {
	          MailHub = strdup (p);
	      	  PortNum = 0;
		  if ((r = strtok (NULL, " \t\r\n:")) != NULL) PortNum = atoi(r);
	          if (PortNum == 0) PortNum = PORTNUMBER;
	          if (LogLevel > 0)
		    {
		      log_event (LOG_INFO,
			         "Set MailHub=\"%s\".\n", MailHub);
		      log_event (LOG_INFO,
			         "via SMTP Port Number=\"%d\".\n", PortNum);
		    }
	        }
	    }
	}
    }
}

/*
 * Process and record headers.
 */

/*
 * recordRequiredHeaders -- note which ones we've seen.
 */
int
recordRequiredHeaders (char *line)
{

  static int	HasBcc = 0,
  		previous_recipient_header = 0;
  
  if (*line == ' ' || *line == '\t')
    {
      if (previous_recipient_header)
	{
	  char	*s = line;
#ifdef DEBUG
fprintf (stderr, "recordRequiredHeaders(): Recipient continuation line \"%s\"\n",
  line);
#endif
	  for (s++; *s == ' ' || *s == '\t'; s++);
	  recordRecipient (s);
	  /* Keep previous_recipient_header set to non-zero. */
	}

      return HasBcc;
    }

  previous_recipient_header = HasBcc = 0;

  /* Better check for ": "-- it's part of the RFC.  -Jeff */
  if (strncasecmp (line, "From: ", 6) == 0) hasFrom = YES;
#ifdef HASTO_OPTION
  else if (strncasecmp(line,"To: ",4)==0) hasTo = YES;
#endif
  else if (strncasecmp (line, "Date: ", 6) == 0) hasDate = YES;
  else if (strncasecmp (line, "Return-Receipt-To: ", 19) == 0) hasDSN = YES;
  else if (Toption)
    {	/* Need to figure out recipients from the e-mail */
      if (strncasecmp(line,"To: ",4)==0)
	{
#ifdef DEBUG
          fprintf (stderr, "recordRequiredHeaders(): Recipient line \"%s\"\n", line);
#endif
	  recordRecipient (line + 4);
	  previous_recipient_header = 1;
	}
      else if (strncasecmp (line, "Bcc: ", 5) == 0)
	{
#ifdef DEBUG
          fprintf (stderr, "recordRequiredHeaders(): Recipient line \"%s\"\n", line);
#endif
	  recordRecipient (line + 5);
	  previous_recipient_header = 1;
	  HasBcc = 1;
	}
      else if (strncasecmp (line, "CC: ", 4) == 0)
	{
#ifdef DEBUG
          fprintf (stderr, "recordRequiredHeaders(): Recipient line \"%s\"\n", line);
#endif
	  recordRecipient (line + 4);
	  previous_recipient_header = 1;
	}
      }
    return HasBcc;
}


/*
 * Extract recipients' email addresses from headers.
 */
void 
recordRecipient (char *line)
{
  char buf[MAXLINE],
       *p;
  struct recips *recip_ent = NULL;
  static struct recips *recip_last;

#ifdef DEBUG
  fprintf (stderr, "recordRecipient(): parsing header line \"%s\"\n", line);
#endif
  for (p = line; *p; p++)
    {
      if ( *p == '"' )
      {
	/* Skip quotes (condensed quotes()) -Jeff */
	for (p++; *p && *p != '"'; p++)
	  if (*p == '\\' && *(p+1))
	    p++;
      }
      else if (*p == ',')
      {
        *p = '\0';
        parseaddr( line, buf, MAXLINE );
#ifdef DEBUG
        fprintf (stderr, "comma separator-- parseaddr returns \"%s\"\n", buf);
#endif
        *p = ',';
        recip_ent = malloc (sizeof *recip_ent);
        recip_ent->next = NULL;
        recip_ent->rec = strdup (buf);
        if (recipients)
          recip_last->next = recip_ent;
        else
          recipients = recip_ent;
        recip_last = recip_ent;
        line = p+1;
      }
    }
  if (*line)
  {
    parseaddr( line, buf, MAXLINE );
#ifdef DEBUG
    fprintf (stderr, "parseaddr returns \"%s\"\n", buf);
#endif
    recip_ent = malloc (sizeof *recip_ent);
    recip_ent->next = NULL;
    recip_ent->rec = strdup (buf);
    if (recipients)
      recip_last->next = recip_ent;
    else
      recipients = recip_ent;
    recip_last = recip_ent;
  }
}

/* 
 * stripFromLine -- transforms "Name <login@host>" into "login@host"
 *                  or "login@host (Real name)"
 */
char *
stripFromLine(char *buffer) 
{
  static char fLine[MAXLINE];
  char *c,*p;

  if(!strncmp("From: ", buffer, 6)) buffer += 6;

  strncpy(fLine, buffer, MAXLINE);

  /* Remove the real name if necessary - just send the address */
  c = strchr (fLine, '<');
  if (c)
    {
      c++;
      p = strchr (c, '>');
      if (p)
	*p = '\0';
    }
  else {
    c = fLine;
    p = strchr(fLine, ' ');
    if (p)
      *p = '\0';
  }

  return c;
}

/*
 * addRequiredHeaders -- add ones that have been missed.
 */
void 
addRequiredHeaders (int fd)
{

  if (hasFrom == NO)
    {
      /* No From: line; add the sender specified in the command line or
	 generate one */
      if (specifiedName == NULL)
        putToSmtp (fd, "From: %s", (specifiedFrom!=NULL) ? specifiedFrom : fromLine);
      else
        putToSmtp (fd, "From: %s <%s>", specifiedName, (specifiedFrom!=NULL) ? specifiedFrom : stripFromLine(fromLine));
    }
#ifdef HASTO_OPTION
  if (hasTo == NO)
    {
      putToSmtp(fd,"To: postmaster");
    }
#endif
  if (hasDate == NO)
    {
      putToSmtp (fd, "Date: %s", DateString);
    }
}

/*
 * addInitialHeaders -- prepend prerequisite timstamp
 *      and actual date lines.
 */
void 
addInitialHeaders (int fd)
{

  putToSmtp (fd, "Received: by %s (sSMTP sendmail emulation); %s",
	     HostName, DateString);

}

#ifdef REWRITE_DOMAIN
/*
 * fixFromLine -- replace whole From: header with standardized pattern.
 *      Evil, nasty, immoral header-rewriting code (:-)).
 */
int 
fixFromLines (char *line)
{
  static int inHeaders = YES;

  if (strncasecmp (line, "From: ", 6) == 0)
    {
      if (FromLineOverride)
        msgFromLine = strdup(line);
      else
        if(snprintf (line, MAXLINE, "From: %s", fromLine) == -1) {
	  die("buffer overflown in fixFromLines");
	}
    }
  if (*line == (char) NULL)
    {
      inHeaders = NO;
    }
  return inHeaders;
}
#endif

/*
 * properRecipient -- alias systems-level users to the person who
 *      reads their mail.  This is variously the owner of a workstation,
 *      the sysadmin of a group of stations and the postmaster otherwise.
 *      We don't just mail stuff off to root on the mailhub (:-)).
 */
char *
properRecipient (char *s)
{
  struct passwd *p;
#ifdef REWRITE_DOMAIN
  static char buffer[MAXLINE];
#endif

  if (!strchr (s, '@')
      && (p = getpwnam (s))
      && p->pw_uid <= 10)
        /* It's a local systems-level user. */
        s = Root;

/* Fix up address with @domain.name. (Might as well inline this one too. -Jeff) */

#ifdef REWRITE_DOMAIN
	if(!strchr(s, '@'))
	{
		if(snprintf(buffer, MAXLINE, "%s@%s",
		s, UseRD == YES ? RewriteDomain : HostName) == -1)
			die("Buffer overflown in fromLine");
		else
			return(buffer);
	}
#endif
	return(s);
}

/*
 * localFromLine -- generate a from line in standard format
 * ("Real Name <id@site>" or "id@site") based on local user and host names.
 */
void
localFromLine (void)
{
  static char buffer[MAXLINE];
  if (fullName) {
    if(snprintf (buffer, MAXLINE, "%s <%s@%s>",
                    fullName,
		    Sender->pw_name,
#ifdef REWRITE_DOMAIN
		    UseRD == YES ? RewriteDomain : HostName
#else
		    HostName
#endif
    		    ) == -1) {
      die("Buffer overflown in fromLine");
    }
  }
  else {
    if(snprintf (buffer, MAXLINE, "%s@%s",
		    Sender->pw_name,
#ifdef REWRITE_DOMAIN
		    UseRD == YES ? RewriteDomain : HostName
#else
		    HostName
#endif
		    ) == -1) {
	die("Buffer overflown in fromLine");
      }
  }
#ifdef DEBUG
  fprintf(stderr,"Fromline generated: %s (%s)\n",buffer,RewriteDomain);
#endif
  fromLine = strdup (buffer);
}

/*
 * Supporting libraries -- signals
 */

/*
 * alarmHandler -- a ``normal'' non-portable version of an alarm handler.
 *      Alas, setting a flag and returning is not fully functional in
 *      BSD: system calls don't fail when reading from a ``slow'' device
 *      like a socket. So we longjump instead, which is erronious on
 *      a small number of machines and ill-defined in the language.
 */
void 
alarmHandler (void)
{
  extern jmp_buf TimeoutJmpBuf;
  longjmp (TimeoutJmpBuf, (int) 1);
}

/*
 * emergency exit functions.
 */

/*
 * flush -- save stdin to dead.letter, if you can.
 */
void 
flush (void)
{
  char line[MAXLINE];
  FILE *fp;

  if (isatty (fileno (stdin)))
    {
      if (LogLevel > 0)
	{
	  log_event (LOG_ERR, "stdin appears to be a terminal. Not saving to dead.letter.");
	}
      return;
    }
  if (Sender == NULL)
    {
      /* Far to early to save things. */
      if (LogLevel > 0)
	{
	  log_event (LOG_ERR, "No sender (can't happen), failing horribly.");
	}
      return;
    }
  if(snprintf (line, MAXLINE, "%s/dead.letter", Sender->pw_dir) == -1) {
    /* can't use die() here since flush() is called from die() */
    exit(1);
  }
  if ((fp = fopen (line, "a")) == NULL)
    {
      /* Perhaps the person doesn't have a homedir... */
      if (LogLevel > 0)
	{
	  log_event (LOG_ERR, "Can't open %s, failing horribly.",
		     line);
	}
      return;
    }
  fprintf (stderr, "%s: sending mail to dead.letter\n", ProgName);
  (void) putc ('\n', fp);	/* Make sure we start on a new line, */
  (void) putc ('\n', fp);	/* with a blank line separating messages. */
  while (fgets (line, sizeof line, stdin))
    {
      (void) fputs (line, fp);
    }
  if (fclose (fp) == ERR)
    {
      if (LogLevel > 0)
	{
	  log_event (LOG_ERR,
		     "Can't close %s/dead.letter, possibly truncated.",
		     Sender->pw_dir);
	}
    }
}

/*
 * Reporting and logging library functions 
 */

/*
 * log_event -- log something to syslog and the log file.
 */
/*VARARGS */
void
log_event (int code, char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  vlog_event (code, fmt, args);
  va_end (args);
}

void 
vlog_event (int syslog_code, char *format, va_list args)
{
#ifdef SYSLOG
  static int syslogOpen = NO;
#endif
#ifdef LOGFILE
  FILE *fp;

  if ((fp = fopen ("/tmp/ssmtp.log", "a")) != NULL)
    {
      (void) vfprintf (fp, format, args);
      (void) putc ('\n', fp);
      (void) fclose (fp);
    }
  else
    {
      /* oops! */
      (void) fprintf (stderr, "Can't write to /tmp/ssmtp.log\n");
    }
#endif
#ifdef SYSLOG
  if (syslogOpen == NO)
    {
      syslogOpen = YES;
#ifdef OLDSYSLOG
      openlog ("sSMTP mail", LOG_PID);
#else
      openlog ("sSMTP mail", LOG_PID, LOG_MAIL);
#endif
    }
  (void) vsyslog (syslog_code, format, args);
#endif
}

/*
 * die -- say something and exit with a non-zero return code.
 *      Save the message on stdin in dead.letter.
 */
void 
die (char *format, ...)
{
  va_list args;
  va_start (args, format);
  (void) fprintf (stderr, "%s: ", ProgName);
  (void) vfprintf (stderr, format, args);
  (void) putc ('\n', stderr);
  flush ();			/* Send message to dead.letter */
  vlog_event (LOG_ERR, format, args);
  va_end (args);
  exit (1);
}

/*
 * Local/peculiar string manipulation.
 */

/*
 * standardize -- trim off '(\r)?\n's, double leading dots.
 */
void 
standardize (char *p)
{
  char *q;

  if ((*p == '.') && (*(p + 1) == '\n'))
    {
      /* Double it, in hopes smtp will single it. */
      *(p + 1) = '.';
      return;
    }

  if ((q = strchr (p, '\n'))) *q = 0;
  else die ("no line terminator at end of header: \"%s\"\n", p);
  if (*--q == '\r') *q = 0;
}

/*
 * Supporting libraries -- i/o.
 */

/*
 * putToSmtp -- a printf to an fd, which appends TCP/IP <CR/LF>.
 */
void 
putToSmtp (int fd, char *format,...)
{
  va_list args;
  char line[MAXLINE];
  va_start (args, format);
  if(vsnprintf (line, MAXLINE-3, format, args) == -1) {
    die("buffer overflown in putToSmtp"); 
  }
  if (LogLevel > 0)
    {
      log_event (LOG_INFO, "Sent \"%s\" to smtp port.\n", line);
    }
#ifdef DEBUG
  printf("%s\n",line);
#endif
  (void) strcat (line, "\r\n");
  (void) write (fd, line, strlen (line));
  va_end (args);
}

/*
 * getOkFromSmtp -- get a line and test the three-number string
 *      at the beginning.  If it starts with a 2, it's OK.
 */
int 
getOkFromSmtp (int fd, char *response)
{
  return (getFromSmtp (fd, response) == 2) ? YES : NO;
}


/*
 * getFromSmtp -- get a line and return the initial digit.  Deal with
 *      continuation lines by reading to the last (non-continuation) line.
 */
int 
getFromSmtp (int fd, char *response)
{
  char *getLine ();

  do
    {
      if (getLine (response, MAXLINE, fd) == NULL)
	{
	  *response = '\0';
	  return NO;
	}
    }
  while (response[3] == '-');
  if (LogLevel > 0)
    {
      log_event (LOG_INFO, "Received \"%s\" from smtp port.\n", response);
    }
  return atoi (response) / 100;
}

/*
 * getLine -- get a line of text from a fd instead of an fp.
 */
char *
getLine (char *line, int size, int fd)
{
  int i;
  char ch;

  for (i = 0; read (fd, &ch, 1) == 1;)
    {
      if (i == size - 1)
	{
	  /* Truncate like fgets. */
	  line[i] = '\0';
	  return line;
	}
      else if (ch == '\r')
	;			/* Strip it. */
      else if (ch == '\n')
	{
	  break;
	}
      else
	{
	  line[i++] = ch;
#ifdef DEBUG
	  line[i] = '\0';
#endif
	}
    }
  line[i] = '\0';
  return line;
}

/* 
 *Code borrowed or adapted from other sources.
 */

/*
 * openSocket -- open a socket on a specified machine.
 *      Adapted from code by Blair P. Houghton:
 *      Copyright 1991 Blair P. Houghton, All Rights Reserved, 
 *      copying and distribution permitted with copyright intact.
 */
int 
openSocket (char *hostName, int portNumber)
{
  int fd;			/* socket to "plug" into the socket */
  struct sockaddr_in socketname;	/* mode, addr, and port data for */
  /* the socket */
  struct hostent *remote_host;	/* internet numbers, names */

  if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    {
      log_event (LOG_ERR, "unable to create a socket.\n");
      return ERR;
    }

  /* plug it into the listening socket */
  socketname.sin_family = AF_INET;
  if ((remote_host = gethostbyname (hostName)) == (struct hostent *) NULL)
    {
      log_event (LOG_ERR, "unable to locate host %s.\n", hostName);
      return ERR;
    }
  (void) bcopy ((char *) remote_host->h_addr, (char *) &socketname.sin_addr,
		remote_host->h_length);
  socketname.sin_port = htons (portNumber);

  if (connect (fd, (struct sockaddr *) &socketname, sizeof socketname) < 0)
    {
      log_event (LOG_ERR, "unable to connect to \"%s\" port %d.\n",
		 hostName, portNumber);
      return ERR;
    }
  return fd;
}

/*
 * parseaddr.c	Read a valid RFC822 address with all the comments
 *		etc in it, and return _just_ the email address.
 *
 * Version:	@(#)parseaddr.c  1.00  02-Apr-1999  miquels@cistron.nl
 *
 */

/*
 *	Return the next token. A token can be "<>()," or any "word".
 */
struct token *gettoken(char **ptr)
{
	struct token	*tok;
	char		*p = *ptr;
	char		*begin;
	int		l, quit = 0;

	SKIPSPACE(p);
	begin = p;

	while (!quit) {
		switch (*p) {
			case 0:
			case ' ':
			case '\t':
			case '\n':
				quit = 1;
				break;
			case '(':
			case ')':
			case '<':
			case '>':
			case ',':
				if (p == begin) p++;
				quit = 1;
				break;
			case '\\':
				if (*(p+1)) p++;
				break;
			case '"':
			/* Skip quotes (condensed quotes()) -Jeff */
				for (p++; *p && *p != '"'; p++)
					if (*p == '\\' && *(p+1))
						p++;
				break;
		}
		if (!quit) p++;
	}

	l = p - begin;
	if (l == 0) return NULL;
	if ((tok = malloc(sizeof(struct token) + l)) == NULL)
		return NULL;
	tok->next = NULL;
	strncpy(tok->word, begin, l);
	tok->word[l] = 0;

	SKIPSPACE(p);
	*ptr = p;

	return tok;
}

/*
 *	Get email address from rfc822 address.
 */
int parseaddr(char *addr, char *buf, int bufsz)
{
	char		*p;
	struct token	*t, *tok, *last;
	struct token	*brace = NULL;
#ifdef DEBUG
	struct token	*dbug = NULL;
#endif
	int		comment = 0;

	tok = last = NULL;

	/*
	 *	Read address, remove comments right away.
	 */
	p = addr;
	while ((t = gettoken(&p)) != NULL && t->word[0] != ',') {
		if (t->word[0] == '(' || t->word[0] == ')' || comment) {
			free(t);
			if (t->word[0] == '(')
				comment++;
			if (t->word[0] == ')')
				comment--;
			continue;
		}
		if (t->word[0] == '<')
			brace = t;
		if (tok)
			last->next = t;
		else
			tok = t;
		last = t;
	}
#ifdef DEBUG
	fprintf (stderr, "token list from parseaddr():\n");
	for (dbug = tok; dbug; dbug = dbug->next)
		fprintf (stderr, "\"%s\"\n", dbug->word);
	putc ('\n', stderr);
#endif
	/*
	 *	Put extracted address into "buf"
	 */
	buf[0] = 0;
	t = brace ? brace->next : tok;
	for (; t && t->word[0] != ',' && t->word[0] != '>'; t = t->next) {
		if (strlen(t->word) >= bufsz) {
			for (t = tok; t; t = last) {
				last = t->next;
				free(t);
			}
			return -1;
		}
		bufsz -= strlen(t->word);
		strcat(buf, t->word);
	}

	/*
	 *	Free list of tokens.
	 */
	for (t = tok; t; t = last) {
		last = t->next;
		free(t);
	}

	return 0;
}

/*
 *  arpadate.c - get_arpadate() is a function returning the date in the
 *               ARPANET format (see RFC822 and RFC1123)
 *  Copyright (C) 1998 Hugo Haas
 *  
 *  Inspired by smail source code by Ronald S. Karr and Landon Curt Noll
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

void 
get_arpadate (char *d_string)
{
  static char *week_day[] =
  {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
  static char *month[] =
  {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
   "Aug", "Sep", "Oct", "Nov", "Dec"};
  static char timezone[6];

  time_t current;
  struct tm *date;
  int offset, gmt_yday, gmt_hour, gmt_min, yday;

  /* Get current time */
  (void) time (&current);

  /* Get GMT and then local dates */
  date = gmtime ((const time_t *) &current);
  gmt_yday = date->tm_yday;
  gmt_hour = date->tm_hour;
  gmt_min = date->tm_min;
  date = localtime ((const time_t *) &current);

  /* Calculates offset */

  offset = ((date->tm_hour - gmt_hour) * 60) + (date->tm_min - gmt_min);

  if ((yday = (date->tm_yday - gmt_yday)))
    {
	/* This code is optimized to negative timezones (West of Greenwich) */
	if (yday == -1 ||   /* UTC passed midnight before localtime */
	    yday > 1)       /* UTC passed new year before localtime */
	    offset -= 24 * 60;
	else
	    offset += 24 * 60;
    }

  if (offset >= 0)
    sprintf (timezone, "+%02d%02d", offset / 60, offset % 60);
  else
    sprintf (timezone, "-%02d%02d", -offset / 60, -offset % 60);

  sprintf (d_string, "%s, %d %s %04d %02d:%02d:%02d %s",
	   week_day[date->tm_wday],
	   date->tm_mday, month[date->tm_mon], date->tm_year + 1900,
	   date->tm_hour, date->tm_min, date->tm_sec, timezone);

}
