#include <stdio.h>
#include <limits.h>
#include <ctype.h>
#include <pwd.h>		/* For getpwent. */
#include <time.h>
#include <sys/types.h>		/* For sockets. */
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#ifdef SYSLOG
#include <syslog.h>		/* For logging. */
#else
#define LOG_ERR 0
#define LOG_INFO 0
#endif
#include <signal.h>		/* For the timer and signals. */
#include <setjmp.h>
#include <string.h>

#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>

   /* #include "string_ext.h"		 Local additions. */
#include "patchlevel.h"

/*
 * spop.h -- Sizing and return-code defines, etc. Stuff you
 *	don't want to think about in the main program.
 */
#define YES   1
#define NO    0 
#define ERR   (-1)

#define MAXLINE	(1024*2) /* A pretty large buffer, but not outrageous. */
#define MAXWAIT (10*60)  /* Maximum wait between commands, in seconds. */
#define MEDWAIT (5*60)
#define ARPADATE_LENGTH	32


#ifndef _POSIX_ARG_MAX
#define MAXARGS 4096
#else
#define MAXARGS  _POSIX_ARG_MAX
#endif
#define CHUNK 4000		/* Sizeof allocation */

#ifdef lint
extern char *sprintf();	/* Ah yes, backward compatability with errors... */
extern void	exit();
#endif

#ifndef MAILHUB
#define MAILHUB "mailhost"	/* A surprisingly useful default. */
#endif
#define PORTNUMBER 25		/* From the Assigned Numbers RFC. */

char *Version = VERSION;	/* The version of the program. */
char *ProgName = NULL;		/* It's name. */
char *MailHub = MAILHUB;	/* The place to send the mail. */
int PortNum = 0;	        /* The sysadmin assigned smtp port number. */
char HostName[MAXLINE];		/* Our name, unless overridden. */
#ifdef REWRITE_DOMAIN
char *RewriteDomain = "localhost";	/* Place to claim to be. */
int UseRD = NO;			/* Do we have to rewrite the domain? */
#endif
int FromLineOverride = NO;      /* Should we use same address in the
                                   from-line of the envelope as in the
                                   From:-line? */
char *msgFromLine = NULL;       /* From:-line from message */
char *fromLine = NULL;          /* From:-line we actually use */
char *specifiedFrom = NULL;	/* Content of the From: field if specified
				 with the -f option */
char *specifiedName = NULL;	/* Same for -F option */
char *fullName = NULL;		/* Sending user's full name */
char *Root = "postmaster";	/* Person to send root's mail to. */
struct passwd *Sender = NULL;	/* The person sending the mail. */
jmp_buf TimeoutJmpBuf;		/* Timeout waiting for input from network. */
int Verbose = NO;		/* Tell the user what's happening. */
int Toption = NO;		/* Was a T option given? */
int LogLevel =			/* Tell the log what's happening. */
#ifdef DEBUG
1;
#else
0;
#endif

char *authUsername = NULL;      /* user name for SMTP authentication */
char *authPassword = NULL;      /* password for SMTP authentication */
int coded_auth;			/* username and password encoded in config file */

/* For list of recipients if -t specified, and for headers */
struct recips {
	struct recips *next;
	char *rec;
}
	*recipients = NULL,
	*headers = NULL;

/*
 * parseaddr.c	Read a valid RFC822 address with all the comments
 *		etc in it, and return _just_ the email address.
 *
 * Version:	@(#)parseaddr.c  1.00  02-Apr-1999  miquels@cistron.nl
 *
 */

struct token {
	struct token *next;
	char word[1];
};

#define SKIPSPACE(p) do { while(*p && isspace(*p)) p++; } while(0)

struct token *gettoken(char **);
int parseaddr(char *, char *, int);


/* Current date in RFC format. */
char DateString[ARPADATE_LENGTH];

#if 0
char *fromLine (void);
#endif
char *properRecipient (char *);

#ifdef REWRITE_DOMAIN
int fixFromLines (char *);
#else
#define fixFromLines(buffer) NO
#endif

void to64frombits(unsigned char *, const unsigned char *, int);
int from64tobits(char *, const char *);
int openSocket (char *, int);
int getOkFromSmtp (int, char *);
int getFromSmtp (int, char *);
void log_event (int, char *, ...);
void vlog_event (int, char *, va_list);
void flush (void);
void die (char *, ...);
void putToSmtp (int, char *,...);
void recordRecipient (char *);
int recordRequiredHeaders (char *);
char *stripFromLine(char *);
void addRequiredHeaders (int);
void addInitialHeaders (int);
void localFromLine (void);
void alarmHandler (void);
void standardize (char *);
void parseConfig (FILE *);
int getConfig (void);
void getReverseAliases (void);
char **doOptions (int, char **);
int from64tobits(char *, const char *);
int openSocket (char *, int);
int getOkFromSmtp (int, char *);
int getFromSmtp (int, char *);
char *getLine (char *, int, int);
void get_arpadate (char *);
